"""retry when 429 is received

   Retries after waiting for the time read from the retry-after header
"""

from functools import wraps
from logging import getLogger
from time import sleep

from .errors import ArchivistTooManyRequestsError

NO_OF_RETRIES = 3
LOGGER = getLogger(__name__)


def retry_429(f):
    """
    Retry when 429 received using sleep suggested by retry_after header
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        no_of_retries = NO_OF_RETRIES
        while True:
            try:
                ret = f(*args, **kwargs)
            except ArchivistTooManyRequestsError as ex:
                if ex.retry <= 0 or no_of_retries <= 0:
                    raise
                sleep(ex.retry)
                no_of_retries -= 1

            else:
                return ret

    return wrapper



"""Locations interface

   Direct access to the locations endpoint.

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      location = arch.locations.create(...)


"""

from contextlib import suppress
from copy import deepcopy
from logging import getLogger
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    # pylint:disable=cyclic-import      # but pylint doesn't understand this feature
    from .archivist import Archivist

from .constants import LOCATIONS_LABEL, LOCATIONS_SUBPATH
from .dictmerge import _deepmerge
from .errors import ArchivistNotFoundError
from .utils import selector_signature

LOGGER = getLogger(__name__)


class Location(dict):
    """Location

    Location object has dictionary attributes.

    """

    @property
    def name(self) -> "str | None":
        """str: name of the location"""
        name = None
        with suppress(KeyError):
            name = self["display_name"]

        return name


class _LocationsClient:
    """LocationsClient

    Access to locations entities using CRUD interface. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._subpath = f"{archivist_instance.root}/{LOCATIONS_SUBPATH}"
        self._label = f"{self._subpath}/{LOCATIONS_LABEL}"

    def __str__(self) -> str:
        return f"LocationsClient({self._archivist.url})"

    def create(
        self, props: "dict[str, Any]", *, attrs: "dict[str, Any]|None" = None
    ) -> Location:
        """Create location

        Creates location with defined properties and attributes.

        Args:
            props (dict): properties for this location.
            attrs (dict): attributes of created location.

        Returns:
            :class:`Location` instance

        """
        LOGGER.debug("Create Location %s", props)
        return self.create_from_data(self.__params(props, attrs))

    def create_from_data(self, data: "dict[str, Any]") -> Location:
        """Create location

        Creates location with request body from data stream.
        Suitable for reading data from a file using json.load or yaml.load

        Args:
            data (dict): request body of location.

        Returns:
            :class:`Location` instance

        """
        return Location(**self._archivist.post(self._label, data))

    def create_if_not_exists(self, data: "dict[str, Any]") -> "tuple[Location, bool]":
        """
        Create a location if not already exists

        Args:
            data (dict): request body of location.

        A YAML representation of the data argument would be:

            .. code-block:: yaml

                selector:
                  - display_name
                  - attributes:
                    - wavestone_ext
                display_name: Apartements du Gare du Nord
                description: Residential apartment building in new complex above GdN station
                latitude: 48.8809
                longitude: 2.3553
                attributes:
                  address: 18 Rue de Dunkerque, 75010 Paris, France
                  wavestone_ext: managed

            The 'selector' setting is required.

        Returns:
            tuple of :class:`Location` instance, Boolean True if already exists

        """

        data = deepcopy(data)
        selector = data.pop("selector")  # must exist
        props, attrs = selector_signature(selector, data)
        try:
            location = self.read_by_signature(props=props, attrs=attrs)

        except ArchivistNotFoundError:
            LOGGER.info(
                "location with selector %s,%s does not exist - creating", props, attrs
            )

        else:
            LOGGER.info("location with selector %s,%s already exists", props, attrs)
            return location, True

        return self.create_from_data(data), False

    def read(self, identity: str) -> Location:
        """Read location

        Reads location.

        Args:
            identity (str): location identity e.g. locations/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            :class:`Location` instance

        """
        return Location(**self._archivist.get(f"{self._subpath}/{identity}"))

    def __params(
        self, props: "dict[str, Any]|None", attrs: "dict[str, Any]|None"
    ) -> "dict[str, Any]":
        params = deepcopy(props) if props else {}
        if attrs:
            params["attributes"] = attrs

        return _deepmerge(self._archivist.fixtures.get(LOCATIONS_LABEL), params)

    def count(
        self,
        *,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
    ) -> int:
        """Count locations.

        Counts number of locations that match criteria.

        Args:
            props (dict): e.g. {"display_name": "Macclesfield" }
            attrs (dict): e.g. {"director": "john smith" }

        Returns:
            integer count of locations.

        """
        return self._archivist.count(self._label, params=self.__params(props, attrs))

    def list(
        self,
        *,
        page_size: "int|None" = None,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
    ):
        """List locations.

        Lists locations that match criteria.

        Args:
            props (dict): optional e.g. {"display_name": "Macclesfield" }
            attrs (dict): optional e.g. {"director": "john smith" }
            page_size (int): optional page size. (Rarely used)

        Returns:
            iterable that returns :class:`Location` instances

        """

        return (
            Location(**a)
            for a in self._archivist.list(
                self._label,
                LOCATIONS_LABEL,
                page_size=page_size,
                params=self.__params(props, attrs),
            )
        )

    def read_by_signature(
        self,
        *,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
    ) -> Location:
        """Read location by signature.

        Reads location that meets criteria. Only one location is expected.

        Args:
            props (dict): e.g. {"display_name": "Macclesfield" }
            attrs (dict): e.g. {"director": "john smith" }

        Returns:
            :class:`Location` instance

        """
        return Location(
            **self._archivist.get_by_signature(
                self._label,
                LOCATIONS_LABEL,
                params=self.__params(props, attrs),
            )
        )



"""Compliance interface

   Access to the compliance endpoint.

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      asset = arch.compliance.compliant_at(...)

"""

from logging import getLogger
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    # pylint:disable=cyclic-import      # but pylint doesn't understand this feature
    from .archivist import Archivist

from .constants import (
    COMPLIANCE_LABEL,
    COMPLIANCE_SUBPATH,
)

LOGGER = getLogger(__name__)


class Compliance(dict):
    """Compliance

    Compliance object has dictionary of all the compliance attributes.

    """


# pylint: disable=too-few-public-methods
class _ComplianceClient:  # pylint: disable=too-few-public-methods
    """ComplianceClient

    Access to compliance entities using CRUD interface. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._subpath = f"{archivist_instance.root}/{COMPLIANCE_SUBPATH}"
        self._label = f"{self._subpath}/{COMPLIANCE_LABEL}"

    def __str__(self) -> str:
        return f"ComplianceClient({self._archivist.url})"

    def compliant_at(
        self,
        asset_id,
        *,
        compliant_at: "bool|None" = None,
        report: "str|None" = None,
    ) -> Compliance:
        """
        Reads compliance of a particular asset.

        Args:
            asset_id (str): asset identity e.g. assets/xxxxxxxxxxxxxxxxxxxxxxx
            compliant_at (str): datetime to check compliance at a particular time (optional).
                                format: rfc3339 - UTC only
                                https://datatracker.ietf.org/doc/html/rfc3339#section-4.1
            report (bool): if true output report
            page_size (int): optional page size. (Rarely used).

        Returns:
            :class:`Compliance` instance

        """
        params = {"compliant_at": compliant_at} if compliant_at is not None else None
        response = self._archivist.get(
            f"{self._label}/{asset_id}",
            params=params,
        )
        if report is True:
            self.compliant_at_report(response)
        return Compliance(**response)

    def compliant_at_report(self, compliance: "dict[str, Any]"):
        """
        Prints report of compliance_at request

        Args:
            compliance (dict): compliance object encapsulating response from compliant_at
        """

        LOGGER.info("Compliant %s", compliance["compliant"])
        for outcome in compliance["compliance"]:
            if outcome["compliant"]:
                continue

            # get the compliance policy
            policy = self._archivist.compliance_policies.read(
                outcome["compliance_policy_identity"]
            )

            # print the policy name and the reason
            LOGGER.info(
                "NON-COMPLIANCE -> Policy: %s: Reason %s",
                policy["display_name"],
                outcome["reason"],
            )



# pylint:  disable=missing-docstring


from logging import getLogger
from os import environ
from sys import exit as sys_exit
from typing import TYPE_CHECKING

from pyaml_env import parse_config

# pylint:disable=cyclic-import      # but pylint doesn't understand this feature
from ... import about

if TYPE_CHECKING:
    from ...archivist import Archivist

LOGGER = getLogger(__name__)


def run(arch: "Archivist", args):
    LOGGER.info("Using version %s of datatrails-archivist", about.__version__)
    LOGGER.info("Namespace %s", args.namespace)

    # if namespace is specified on the commandline then override any environment
    # setting...
    if args.namespace:
        environ["DATATRAILS_UNIQUE_ID"] = args.namespace

    with open(args.yamlfile, "r", encoding="utf-8") as yml:
        arch.runner(parse_config(data=yml))

    sys_exit(0)



"""Archivist SDK

   Manage headers allowing for upper/lower/canonicalize form
"""

from requests import models


def _headers_get(headers: models.CaseInsensitiveDict, key: str) -> "str|None":
    if headers is not None:
        ret = headers.get(key)
        if ret is not None:
            return ret
        return headers.get(key.lower())

    return None



"""Composite interface

   Access to various utility methods that access various endpoints.

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      asset = arch.composite.......(...)

"""

from logging import getLogger
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # pylint:disable=cyclic-import      # but pylint doesn't understand this feature
    from .archivist import Archivist


LOGGER = getLogger(__name__)


class _CompositeClient:
    """CompositeClient

    Access to various composite methods,
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    These mthods are not unittested and provided as a convenience.
    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance

    def __str__(self) -> str:
        return f"CompositeClient({self._archivist.url})"

    def estate_info(self):
        """
        Evaluate health of the various assets and events in the system

        The report is emitted using LOGGER.info statements
        """
        num_events = self._archivist.events.count()
        num_assets = self._archivist.assets.count()
        num_locations = self._archivist.locations.count()

        LOGGER.info(
            (
                "There are %s events registered against %s assets"
                " in the system spread over %s locations."
            ),
            num_events,
            num_assets,
            num_locations,
        )



"""Events interface

   Direct access to the events endpoint.

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      asset = arch.assets.create(...)
      event = arch.events.create(asset['identity'], ...)

"""

from copy import deepcopy
from logging import getLogger
from typing import TYPE_CHECKING, Any

# pylint:disable=cyclic-import      # but pylint doesn't understand this feature
from . import confirmer
from .constants import (
    ASSETS_SUBPATH,
    ASSETS_WILDCARD,
    CONFIRMATION_STATUS,
    EVENTS_LABEL,
    SBOM_RELEASE,
)
from .dictmerge import _deepmerge
from .errors import ArchivistBadFieldError, ArchivistNotFoundError
from .sboms import sboms_parse

if TYPE_CHECKING:
    from .archivist import Archivist

LOGGER = getLogger(__name__)


class Event(dict):
    """Event

    Event object has dictionary attributes and properties.

    """

    @property
    def when(self) -> "str | None":
        """when

        Timestamp of event

        """
        try:
            when: str = self["timestamp_declared"]
        except KeyError:
            pass
        else:
            return when

        try:
            when: str = self["timestamp_accepted"]
        except KeyError:
            pass
        else:
            return when

        return None

    @property
    def who(self) -> "str | None":
        """who

        Principal identity.

        """

        try:
            who: str = self["principal_declared"]["display_name"]
        except (KeyError, TypeError):
            pass
        else:
            return who

        try:
            who: str = self["principal_accepted"]["display_name"]
        except (KeyError, TypeError):
            pass
        else:
            return who

        return None


class _EventsPublic:
    """EventsPublic

    Access to events entities using the CRUD interface. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._public = archivist_instance.public
        self._subpath = f"{archivist_instance.root}/{ASSETS_SUBPATH}"

    def __str__(self) -> str:
        return "EventsPublic()"

    def _identity(self, identity: str) -> str:
        """Return fully qualified identity
        If public then expect a full url as argument
        """

        if self._public:
            return identity

        return f"{self._subpath}/{identity}"

    def read(self, identity: str) -> Event:
        """Read event

        Reads event.

        Args:
            identity (str): events identity e.g. assets/xxxxxxx.../events/yyyyyyy...

        Returns:
            :class:`Event` instance

        """
        return Event(**self._archivist.get(f"{self._identity(identity)}"))

    def _params(
        self,
        props: "dict[str, Any]|None",
        attrs: "dict[str, Any]|None",
        asset_attrs: "dict[str, Any]|None",
    ) -> "dict[str, Any]":
        params = deepcopy(props) if props else {}
        if attrs:
            params["event_attributes"] = attrs
        if asset_attrs:
            params["asset_attributes"] = asset_attrs

        return _deepmerge(self._archivist.fixtures.get(EVENTS_LABEL), params)

    def count(
        self,
        *,
        asset_id: "str|None" = None,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
        asset_attrs: "dict[str, Any]|None" = None,
    ) -> int:
        """Count events.

        Counts number of events that match criteria.

        Args:
            asset_id (str): optional asset identity e.g. assets/xxxxxxxxxxxxxxxxxxxxxxxxxx
            props (dict): optional properties e.g. {"confirmation_status": "CONFIRMED" }
            attrs (dict): optional attributes e.g. {"arc_display_type": "open" }
            asset_attrs (dict): optional asset_attributes e.g. {"arc_display_type": "door" }

        Returns:
            integer count of assets.

        """

        # wildcarding not allowed when public - asset_id is required (not optional)
        # if asset_id is wildcarded a 401 will be returned from upstream
        if not self._public and not asset_id:
            asset_id = ASSETS_WILDCARD

        # The type checker rightly points out in the case of an event being public but with no
        # asset_id will cause issues in the _identity function and the count function
        LOGGER.debug("asset_id %s", asset_id)
        LOGGER.debug(
            "event_id %s",
            f"{self._identity(asset_id)}/{EVENTS_LABEL}",  # pyright: ignore
        )
        return self._archivist.count(
            f"{self._identity(asset_id)}/{EVENTS_LABEL}",  # pyright: ignore
            params=self._params(props, attrs, asset_attrs),
        )

    def list(
        self,
        *,
        asset_id: "str|None" = None,
        page_size: "int|None" = None,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
        asset_attrs: "dict[str, Any]|None" = None,
    ):
        """List events.

        Lists events that match criteria.

        Args:
            asset_id (str): optional asset identity e.g. assets/xxxxxxxxxxxxxxxxxxxxxxxxxx
            props (dict): e.g. {"tracked": "TRACKED" }
            attrs (dict): e.g. {"arc_display_type": "open" }
            asset_attrs (dict): optional asset_attributes e.g. {"arc_display_type": "door" }
            page_size (int): optional page size. (Rarely used).

        Returns:
            iterable that returns :class:`Event` instances

        """
        # wildcarding not allowed when public - asset_id is required (not optional)
        # if asset_id is wildcarded a 401 will be returned from upstream
        if not self._public:
            asset_id = asset_id or ASSETS_WILDCARD

        return (
            Event(**a)
            for a in self._archivist.list(
                f"{self._identity(asset_id)}/{EVENTS_LABEL}",  # pyright: ignore
                EVENTS_LABEL,
                page_size=page_size,
                params=self._params(props, attrs, asset_attrs),
            )
        )

    def read_by_signature(
        self,
        *,
        asset_id: "str|None" = None,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
        asset_attrs: "dict[str, Any]|None" = None,
    ) -> Event:
        """Read event by signature.

        Reads event that meets criteria. Only one event is expected.

        Args:
            asset_id (str): optional asset identity e.g. assets/xxxxxxxxxxxxxxxxxxxxxxxxxx
            props (dict): e.g. {"tracked": "TRACKED" }
            attrs (dict): e.g. {"arc_display_type": "open" }
            asset_attrs (dict): optional asset_attributes e.g. {"arc_display_type": "door" }

        Returns:
            :class:`Event` instance

        """
        # wildcarding not allowed when public - asset_id is required (not optional)
        # if asset_id is wildcarded a 401 will be returned from upstream
        if not self._public:
            asset_id = asset_id or ASSETS_WILDCARD

        return Event(
            **self._archivist.get_by_signature(
                f"{self._identity(asset_id)}/{EVENTS_LABEL}",  # pyright: ignore
                EVENTS_LABEL,
                params=self._params(props, attrs, asset_attrs),
            )
        )


class _EventsRestricted(_EventsPublic):
    """EventsRestricted

    Access to events entities using the CRUD interface. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        super().__init__(archivist_instance)
        self.pending_count: int = 0

    def __str__(self) -> str:
        return f"EventsRestricted({self._archivist.url})"

    def create(
        self,
        asset_id: str,
        props: "dict[str, Any]",
        attrs: "dict[str, Any]",
        *,
        asset_attrs: "dict[str, Any]|None" = None,
        confirm: bool = False,
    ) -> Event:
        """Create event

        Creates event for given asset.

        Args:
            asset_id (str): asset identity e.g. assets/xxxxxxxxxxxxxxxxxxxxxxxxxx
            props (dict): properties for this event.
            attrs (dict): attributes of created event.
            asset_attrs (dict): attributes of referenced asset.
            confirm (bool): if True wait for event to be confirmed.

        Returns:
            :class:`Event` instance

        """

        LOGGER.debug("Create Event %s/%s", asset_id, props)
        return self.create_from_data(
            asset_id,
            self._params(props, attrs, asset_attrs),
            confirm=confirm,
        )

    def create_from_data(
        self, asset_id: str, data: "dict[str, Any]", *, confirm: bool = False
    ) -> Event:
        """Create event

        Creates event for given asset from data.
        Suitable for reading data from json.load or yaml.load from a file

        Args:
            asset_id (str): asset identity e.g. assets/xxxxxxxxxxxxxxxxxxxxxxxxxx
            data (dict): request body of event.
            confirm (bool): if True wait for event to be confirmed.

        Returns:
            :class:`Event` instance

        """
        data = deepcopy(data)

        event_attributes = data["event_attributes"]
        # is location present?
        location = data.pop("location", None)
        if location is not None:
            if "identity" in location:
                data["event_attributes"]["arc_location_identity"] = location["identity"]
            else:
                loc, _ = self._archivist.locations.create_if_not_exists(
                    location,
                )
                event_attributes["arc_location_identity"] = loc["identity"]

        attachments = data.pop("attachments", None)
        if attachments is not None:
            for a in attachments:
                result = self._archivist.attachments.create(a)
                if a.get("type") == SBOM_RELEASE:
                    sbom_result = sboms_parse(a)
                    for k, v in sbom_result.items():
                        event_attributes[f"sbom_{k}"] = v

                    event_attributes["sbom_identity"] = result["arc_blob_identity"]

                attachment_key = a.get("attachment", None)
                if attachment_key is None:
                    # failing that create a key from filename or url
                    attachment_key = self._archivist.attachments.get_default_key(a)
                event_attributes[attachment_key] = result

        data["event_attributes"] = event_attributes

        event = Event(
            **self._archivist.post(f"{self._subpath}/{asset_id}/{EVENTS_LABEL}", data)
        )
        if not confirm:
            return event

        event_id: str = event["identity"]
        return self.wait_for_confirmation(event_id)

    def wait_for_confirmation(self, identity: str) -> Event:
        """Wait for event to be confirmed.

        Waits for event to be confirmed.

        Args:
            identity (str): identity of event

        Returns:
            True if event is confirmed.

        """
        confirmer.MAX_TIME = self._archivist.max_time
        # pylint: disable=protected-access
        return confirmer._wait_for_confirmation(self, identity)

    def wait_for_confirmed(
        self,
        *,
        asset_id: "str|None" = None,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
        asset_attrs: "dict[str, Any]|None" = None,
    ) -> bool:
        """Wait for events to be confirmed.

        Waits for all events that match criteria to be confirmed.

        Args:
            asset_id (str): optional asset identity e.g. assets/xxxxxxxxxxxxxxxxxxxxxxxxxx
            props (dict): e.g. {"tracked": "TRACKED" }
            attrs (dict): e.g. {"arc_display_type": "open" }
            asset_attrs (dict): optional asset_attributes e.g. {"arc_display_type": "door" }

        Returns:
            True if all events are confirmed.

        """
        asset_id = asset_id or ASSETS_WILDCARD
        # check that entities exist
        newprops = deepcopy(props) if props else {}
        newprops.pop(CONFIRMATION_STATUS, None)

        LOGGER.debug("Count events %s", newprops)
        count = self.count(
            asset_id=asset_id, props=newprops, attrs=attrs, asset_attrs=asset_attrs
        )
        if count == 0:
            raise ArchivistNotFoundError("No events exist")

        confirmer.MAX_TIME = self._archivist.max_time
        # pylint: disable=protected-access
        return confirmer._wait_for_confirmed(
            self,
            asset_id=asset_id,
            props=props,
            attrs=attrs,
            asset_attrs=asset_attrs,
        )

    def publicurl(self, identity: str) -> str:
        """Read event public url

        Reads event public url.

        Args:
            identity (str): events identity e.g. assets/xxxxxxx.../events/yyyyyyy...

        Returns:
            :str:public url as a string

        """
        body = self._archivist.get(f"{self._identity(identity)}:publicurl")
        publicurl = body.get("publicurl")
        if publicurl is None:
            raise ArchivistBadFieldError("No publicurl found in response")

        return publicurl



"""
Base runner class for interpreting yaml story files.

"""

from collections import defaultdict
from functools import partialmethod
from json import dumps as json_dumps
from logging import getLogger
from time import sleep as time_sleep
from types import GeneratorType
from typing import TYPE_CHECKING, Any, Callable
from uuid import UUID

from .errors import ArchivistError, ArchivistInvalidOperationError

# pylint:disable=cyclic-import      # but pylint doesn't understand this feature
# pylint:disable=missing-function-docstring
# pylint:disable=protected-access
if TYPE_CHECKING:
    from .archivist import Archivist

LOGGER = getLogger(__name__)


NOUNS = ("asset", "location", "subject")


def tree():
    """Recursive dict of dicts"""
    return defaultdict(tree)


class _ActionMap(dict):
    """
    Map of actions and keywords for an action
    """

    # 'use_asset_label' gets the asset_idenity and can insert it in 3 places:
    #
    # 1. first positional argument
    # 2. keyword arguments with key
    # 3. keyword argument in first argument which is a dictionary.
    #
    #  use_asset_label = 1   = first positional argument
    #  use_asset_label = "asset_id"   = keyword argument
    #  use_asset_label = "-asset_id"   = keyword argument in first argumemt that is
    #                                    a dictionary
    # similarly for location and subjects labels
    #
    def __init__(self, archivist_instance: "Archivist"):
        super().__init__()
        self._archivist = archivist_instance

        # please keep in alphabetical order
        self["ASSETS_ATTACHMENT_INFO"] = {
            "action": self._archivist.attachments.info,
            "use_asset_label": "add_arg_identity",
        }
        self["ASSETS_COUNT"] = {
            "action": self._archivist.assets.count,
            "keywords": (
                "props",
                "attrs",
            ),
        }
        self["ASSETS_CREATE_IF_NOT_EXISTS"] = {
            "action": self._archivist.assets.create_if_not_exists,
            "keywords": ("confirm",),
            "set_asset_label": True,
            "use_location_label": "add_data_location_identity",
        }
        self["ASSETS_CREATE"] = {
            "action": self._archivist.assets.create_from_data,
            "keywords": ("confirm",),
            "set_asset_label": True,
        }
        self["ASSETS_LIST"] = {
            "action": self._archivist.assets.list,
            "keywords": (
                "props",
                "attrs",
            ),
        }
        self["ASSETS_WAIT_FOR_CONFIRMED"] = {
            "action": self._archivist.assets.wait_for_confirmed,
            "keywords": (
                "props",
                "attrs",
            ),
        }
        self["COMPOSITE_ESTATE_INFO"] = {
            "action": self._archivist.composite.estate_info,
        }
        self["COMPLIANCE_POLICIES_CREATE"] = {
            "action": self._archivist.compliance_policies.create_from_data,
            "delete": self._archivist.compliance_policies.delete,
        }
        self["COMPLIANCE_COMPLIANT_AT"] = {
            "action": self._archivist.compliance.compliant_at,
            "keywords": ("report",),
            "use_asset_label": "add_arg_identity",
        }
        self["EVENTS_CREATE"] = {
            "action": self._archivist.events.create_from_data,
            "keywords": ("confirm",),
            "use_asset_label": "add_arg_identity",
            "use_location_label": "add_data_location_identity",
        }
        self["EVENTS_COUNT"] = {
            "action": self._archivist.events.count,
            "keywords": (
                "asset_id",
                "props",
                "attrs",
                "asset_attrs",
            ),
            "use_asset_label": "add_kwarg_asset_identity",
        }
        self["EVENTS_LIST"] = {
            "action": self._archivist.events.list,
            "keywords": (
                "asset_id",
                "props",
                "attrs",
                "asset_attrs",
            ),
            "use_asset_label": "add_kwarg_asset_identity",
        }
        self["LOCATIONS_COUNT"] = {
            "action": self._archivist.locations.count,
            "keywords": (
                "props",
                "attrs",
            ),
        }
        self["LOCATIONS_CREATE_IF_NOT_EXISTS"] = {
            "action": self._archivist.locations.create_if_not_exists,
            "keywords": ("confirm",),
            "set_location_label": True,
        }
        self["LOCATIONS_LIST"] = {
            "action": self._archivist.locations.list,
            "keywords": (
                "props",
                "attrs",
            ),
        }
        self["LOCATIONS_READ"] = {
            "action": self._archivist.locations.read,
            "use_location_label": "add_arg_identity",
        }
        self["SUBJECTS_COUNT"] = {
            "action": self._archivist.subjects.count,
            "keywords": ("display_name",),
        }
        self["SUBJECTS_CREATE"] = {
            "action": self._archivist.subjects.create_from_data,
            "delete": self._archivist.subjects.delete,
            "set_subject_label": True,
        }
        self["SUBJECTS_CREATE_FROM_B64"] = {
            "action": self._archivist.subjects.create_from_b64,
            "delete": self._archivist.subjects.delete,
            "set_subject_label": True,
        }
        self["SUBJECTS_DELETE"] = {
            "action": self._archivist.subjects.delete,
            "use_subject_label": "add_arg_identity",
        }
        self["SUBJECTS_LIST"] = {
            "action": self._archivist.subjects.list,
            "keywords": ("display_name",),
        }
        self["SUBJECTS_READ"] = {
            "action": self._archivist.subjects.read,
            "use_subject_label": "add_arg_identity",
        }
        self["SUBJECTS_UPDATE"] = {
            "action": self._archivist.subjects.update,
            "keywords": (
                "display_name",
                "wallet_pub_key",
                "tessera_pub_key",
            ),
            "use_subject_label": "add_arg_identity",
        }
        self["SUBJECTS_WAIT_FOR_CONFIRMATION"] = {
            "action": self._archivist.subjects.wait_for_confirmation,
            "use_subject_label": "add_arg_identity",
        }

    def ops(self, action_name: str) -> "dict[str, Any]":
        """
        Get valid entry in map
        """
        ops = self.get(action_name)
        if ops is None:
            raise ArchivistInvalidOperationError(f"Illegal Action '{action_name}'")
        return ops

    def action(self, action_name: str) -> Callable:
        """
        Get valid action in map
        """
        # if an exception occurs here then the dict initialized above is faulty.
        return self.ops(action_name).get("action")  # pyright: ignore

    def keywords(self, action_name: str) -> "tuple | None":
        """
        Get keywords in map
        """
        return self.ops(action_name).get("keywords")

    def delete(self, action_name: str):
        """
        Get delete_method in map
        """
        return self.ops(action_name).get("delete")

    def label(self, noun: str, endpoint: str, action_name: str) -> bool:
        """
        Return whether this action uses or sets label
        """
        return self.ops(action_name).get(f"{noun}_{endpoint}_label", False)


class _Step(dict):  # pylint:disable=too-many-instance-attributes
    def __init__(self, archivist_instance: "Archivist", **kwargs):
        super().__init__(**kwargs)
        self._archivist = archivist_instance
        self._args: "list[Any]" = []
        self._kwargs: "dict[str, Any]" = {}
        self._actions = None
        self._action = None
        self._action_name = None
        self._data = {}
        self._keywords = None
        self._delete_method = None
        self._labels = {}
        self._labels["use"] = {}
        self._labels["set"] = {}

    def add_arg_identity(self, identity):
        self._args.append(identity)

    def add_kwarg_identity(self, key, identity):
        self._kwargs[key] = identity

    add_kwarg_asset_identity = partialmethod(add_kwarg_identity, "asset_id")

    def add_data_identity(self, key, identity):
        self._data[key] = {}
        self._data[key]["identity"] = identity

    add_data_location_identity = partialmethod(add_data_identity, "location")

    def args(self, identity_method, step):
        """
        Add args and kwargs to action.
        """
        self._args = []
        self._kwargs = {}

        # keys are values that must be removed from the body of the request.
        # These are typically 'confirm' or 'report'. Some of the actions
        # have longer lists of keywords - these actions will be simplified
        # when dataclasses are introduced and this code will be much simpler.
        keys = []
        keywords = self.keywords
        if keywords is not None and len(keywords) > 0:
            keys.extend(keywords)
            for k in keywords:
                if k in step:
                    self._kwargs[k] = step[k]

        # add the request body to the positional arguments...
        self._data = {k: v for k, v in step.items() if k not in keys}

        for noun in NOUNS:
            label = self.get(f"{noun}_label")
            func = self.label("use", noun)
            if label is not None and func:
                identity = self.identity_from_label(noun, identity_method)
                if identity is None:
                    raise ArchivistInvalidOperationError(f"unknown {noun} '{label}'")

                getattr(self, func)(identity)

        if self._data:
            self._args.append(self._data)

    def execute(self):
        action = self.action
        LOGGER.debug("action %s", action)
        LOGGER.debug("args %s", self._args)
        if len(self._kwargs) > 0:
            LOGGER.debug("kwargs %s", self._kwargs)
            response = action(*self._args, **self._kwargs)
        else:
            response = action(*self._args)

        # Some actions return a tuple
        if isinstance(response, tuple):
            response = response[0]

        return response

    def label(self, verb: str, noun: str):
        if self._labels[verb].get(noun) is None:
            self._labels[verb][noun] = self.actions.label(verb, noun, self.action_name)

        return self._labels[verb][noun]

    def identity_from_label(self, noun, identity_method):
        label = self.get(f"{noun}_label")
        if not label.startswith(f"{noun}s/"):  # pyright: ignore
            return identity_method(label)

        uid = label.split("/")[1]  # pyright: ignore
        try:
            _ = UUID(uid, version=4)
        except ValueError:
            return None

        return label

    def description(self):
        description = self.get("description")
        if description is not None:
            LOGGER.info(description)

    @property
    def delete(self):
        return self.get("delete")

    def print_response(self, response):
        print_response = self.get("print_response")
        if print_response:
            # some responses are generators...
            if isinstance(response, GeneratorType):
                for e in response:
                    LOGGER.info("Response %s", json_dumps(e, indent=4))
            else:
                LOGGER.info("Response %s", json_dumps(response, indent=4))

    def wait_time(self):
        wait_time = self.get("wait_time", 0)
        if wait_time > 0:
            LOGGER.info("Waiting for %d seconds", wait_time)
            time_sleep(wait_time)

    @property
    def actions(self):
        if self._actions is None:
            self._actions = _ActionMap(self._archivist)

        return self._actions

    @property
    def action(self) -> Callable:
        if self._action is None:
            self._action = self.actions.action(self.action_name)

        return self._action

    @property
    def delete_method(self):
        if self._delete_method is None:
            self._delete_method = self.actions.delete(self.action_name)

        return self._delete_method

    @property
    def keywords(self):
        if self._keywords is None:
            self._keywords = self.actions.keywords(self.action_name)

        return self._keywords

    @property
    def action_name(self):
        if self._action_name is None:
            action_name = self.get("action")
            if action_name is None:
                raise ArchivistInvalidOperationError("Missing Action")

            self._action_name = action_name

        return self._action_name


class _Runner:
    """
    ArchivistRunner takes a url, token_file.
    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self.entities: defaultdict
        self.deletions = {}

    def __str__(self) -> str:
        return f"Runner({self._archivist.url})"

    def __call__(self, config: "dict[str, Any]"):
        """
        The dict config contains a list of `steps` to be performed serially, e.g.

        ```
        "steps": [
            {
                "step": {
                    "action": "ASSETS_CREATE",
                    "description": "Create a Radiation bag number one",
                    "wait_time": 10,
                },
                "attributes": {
                    "arc_display_name": "radiation bag 1",
                    "radioactive": True,
                    "radiation_level": 0,
                    "weight": 0,
                },
            }
        ]
        ```

        where:
         - `action` is the operation to perform.
         - `description` is what to print to the console.
         - `wait_time` is time to wait before running the operation.
         - `attributes` are the asset's attributes

        To perform all the steps call the class instance.
        """
        try:
            self.run_steps(config)
        except (ArchivistError, KeyError) as ex:
            LOGGER.info("Runner exception %s", ex)

    def run_steps(self, config: "dict[str, Any]"):
        """Runs all defined steps in self.config."""
        self.entities = tree()
        for step in config["steps"]:
            self.run_step(step)

        self.delete()
        self._archivist.close()

    def run_step(self, step: "dict[str, Any]"):
        """Runs a step given parameters and the type of step.

        Args:
            step (dict): the steps map.
        """

        # get step settings
        s = _Step(self._archivist, **step.pop("step"))

        # output description
        s.description()

        # this is a bit clunky...
        s.args(self.identity, step)

        # wait for a number of seconds and then execute
        s.wait_time()
        response = s.execute()

        s.print_response(response)

        if s.delete:
            self.set_deletions(response, s.delete_method)

        for noun in NOUNS:
            label = s.get(f"{noun}_label")
            if s.label("set", noun) and label is not None:
                self.entities[label] = response

    def set_deletions(self, response: "dict[str, Any]", delete_method):
        """sets entry to be deleted"""

        if delete_method is not None:
            identity = response["identity"]
            self.deletions[identity] = delete_method

    def delete(self):
        """Deletes all entities"""
        for identity, delete_method in self.deletions.items():
            LOGGER.info("Delete %s", identity)
            delete_method(identity)

    def identity(self, name: str) -> "str|None":
        """Gets entity id"""

        identity = self.entities[name]["identity"]
        if isinstance(identity, str):
            return identity

        return None



"""Archivist Proof Mechanism

   Enumerated type that allows user to select the proof_mechanism option when
   creating an asset.

   Currently there is only one proof mechanism so this code is here only for
   compatibility.
"""

# pylint: disable=unused-private-member

from enum import Enum


class ProofMechanism(Enum):
    """Enumerate proof mechanism options"""

    # previously used but now removed
    __RESERVED1 = 1
    SIMPLE_HASH = 2
    #: Assets and events are proven using a merkle log hash of the originator's evidence
    MERKLE_LOG = 3



"""Archivist SDK

   WARNING: Do **not** edit directly
   This file is auto-generated by ./scripts/version.sh and is
   not version controlled by git (ironically) as this would
   make the generated version from the tag 'dirty'
"""
__version__ = "v0.32.0"



"""Tenancies interface

   Access to the tenancies endpoint.

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      tenancy = arch.tenancies.read(...)

"""

from logging import getLogger
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # pylint:disable=cyclic-import      # but pylint doesn't understand this feature
    from .archivist import Archivist

from .constants import (
    TENANCIES_LABEL,
    TENANCIES_PREFIX,
    TENANCIES_SUBPATH,
)

LOGGER = getLogger(__name__)


class Tenant(dict):
    """Tenant object"""


class _TenanciesClient:
    """TenanciesClient

    Access to tenants entities using CRUD interface. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    maxDiff = None

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._subpath = f"{archivist_instance.root}/{TENANCIES_SUBPATH}"
        self._label = f"{self._subpath}/{TENANCIES_LABEL}"

    def __str__(self) -> str:
        return f"TenanciesClient({self._archivist.url})"

    def _identity(self, identity: str) -> str:
        """Returns identity suitable for endpoint"""
        prefix, uuid = identity.split("/")
        if prefix == TENANCIES_PREFIX:
            return "/".join((TENANCIES_LABEL, uuid))

        return identity

    def publicinfo(self, identity: str) -> Tenant:
        """Read Tenant public info

        Reads Tenant public info

        Args:
            identity (str): tenancies identity e.g. tenancies/xxxxxxxxxxxxxxxxxxxxxxx
                                                 or tenant/xxxxxxxxxxxxxxxxxxxxxxx

        The tenant identity returned by most endpoints has the prefix 'tenant/'.
        However the tenancies endpoint expects 'tenancies/'. This is a wart in the
        archivist API and will be fixed at some future date.

        Returns:
            :class:`Tenant` instance

        """
        return Tenant(
            **self._archivist.get(
                f"{self._subpath}/{self._identity(identity)}:publicinfo"
            )
        )



"""Archivist Confirmation Status

   Enumerated type that allows user to select the confirmation status option when
   creating an asset.

"""

# pylint: disable=unused-private-member

from enum import Enum


class ConfirmationStatus(Enum):
    """Enumerate confirmation status options"""

    UNSPECIFIED = 0
    # not yet committed
    PENDING = 1
    CONFIRMED = 2
    # permanent failure
    FAILED = 3
    # forestrie, "its in the db"
    STORED = 4
    # forestrie, "you can know if its changed"
    COMMITTED = 5
    # forestrie, "You easily prove it was publicly available to all"
    UNEQUIVOCAL = 6



# pylint:  disable=missing-docstring

from logging import getLogger
from os import environ
from pathlib import PurePath
from sys import exit as sys_exit

import yaml
from jinja2 import Environment, FileSystemLoader

from ... import about

# pylint:disable=cyclic-import      # but pylint doesn't understand this feature
from ... import archivist as type_helper  # pylint:disable=unused-import

LOGGER = getLogger(__name__)


def run(arch: "type_helper.Archivist", args):
    LOGGER.info("Using version %s of datatrails-archivist", about.__version__)
    LOGGER.info("Namespace %s", args.namespace)

    path = PurePath(args.template)
    jinja = Environment(
        loader=FileSystemLoader(path.parent),
        trim_blocks=True,
        lstrip_blocks=True,
    )
    template = jinja.get_template(path.name)

    # if namespace is specified on the commandline then override any environment
    # setting...
    if args.namespace:
        environ["DATATRAILS_UNIQUE_ID"] = args.namespace

    # environment is injected into the template
    with open(args.values, "r", encoding="utf-8") as fd:
        arch.runner(
            yaml.load(
                template.render(
                    yaml.load(
                        fd,
                        Loader=yaml.SafeLoader,
                    ),
                    env=environ,
                ),
                Loader=yaml.SafeLoader,
            ),
        )

    sys_exit(0)



"""Some convenience stuff
"""

from logging import getLogger
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from io import BytesIO

from requests import get as requests_get

LOGGER = getLogger(__name__)


# pylint: disable=missing-function-docstring
def __tuple_member(tup, idx, default):
    try:
        ret = tup[idx]
    except IndexError:
        ret = default

    return ret


# for confirmers
def backoff_handler(details):
    wait = details["wait"]
    tries = details["tries"]
    args = details["args"]
    client = __tuple_member(args, 0, "")
    identity = __tuple_member(args, 1, "")
    LOGGER.debug(
        "Backing off %0.1f seconds after %s tries with %s id %s",
        wait,
        tries,
        client,
        identity,
    )


# download arbitrary files from a url.
def get_url(url: str, fd: "BytesIO"):  # pragma: no cover
    """GET method (REST) - chunked

    Downloads a binary object from upstream storage.
    """
    response = requests_get(
        url,
        stream=True,
        timeout=30,
    )

    for chunk in response.iter_content(chunk_size=4096):
        if chunk:
            fd.write(chunk)


def get_auth(
    *,
    auth_token_filename=None,
    auth_token=None,
    client_id=None,
    client_secret_filename=None,
    client_secret=None,
):  # pragma: no cover
    """
    Return auth as either stuntidp token or client_id,client_secret tuple
    """

    if auth_token:
        return auth_token

    if auth_token_filename:
        with open(auth_token_filename, mode="r", encoding="utf-8") as tokenfile:
            auth_token = tokenfile.read().strip()

        return auth_token

    if client_id is not None:
        if client_secret_filename is not None:
            with open(client_secret_filename, mode="r", encoding="utf-8") as tokenfile:
                client_secret = tokenfile.read().strip()
            return (client_id, client_secret)

        if client_secret is not None:
            return (client_id, client_secret)

    return None


def selector_signature(selector: list, data: dict) -> "tuple[dict, dict | None]":
    """
    Convert a selector to a signature for list and count methods

    Used in locations.create_if_not_exists and assets.create_if_not_exists
    """
    attrselector = []
    propselector = []

    for k in selector:
        try:
            attrselector = k["attributes"]
        except (KeyError, TypeError):
            propselector.append(k)

    props = {k: data[k] for k in propselector}
    if attrselector:
        data_attrs = data["attributes"]  # keyerror if not exist (it must exist)
        attrs = {k: data_attrs[k] for k in attrselector}
    else:
        attrs = None

    return (props, attrs)



"""SBOMS interface

   Convenience methods for handling sbom attachments

"""

# pylint:disable=too-few-public-methods


from io import BytesIO
from logging import getLogger
from typing import Any

from xmltodict import parse as xmltodict_parse

from .utils import get_url

LOGGER = getLogger(__name__)


def sboms_parse(data: "dict[str, Any]") -> "dict[str, Any]":  # pragma: no cover
    """
    parse the sbom and extract pertinent information

    Args:
        data (dict): dictionary

    A YAML representation of the data argument would be:

        .. code-block:: yaml

            filename: functests/test_resources/sboms/gen1.xml

        OR

        .. code-block:: yaml

            url: https://some.hostname/cdx.xml

         Either 'filename' or 'url' is required.

    Returns:

        A dict suitable for adding to an asset or event creation

    """
    result = None
    filename = data.get("filename")
    if filename is not None:
        with open(filename, "rb") as fd:
            sbom = xmltodict_parse(fd, xml_attribs=True, disable_entities=False)

    else:
        url = data["url"]
        fd = BytesIO()
        get_url(url, fd)
        sbom = xmltodict_parse(fd, xml_attribs=True, disable_entities=False)

    b = sbom["bom"]
    m = b["metadata"]
    c = m["component"]

    result = {
        "author": c["author"],
        "component": c["name"],
        "supplier": c["supplier"]["name"],
        "version": c["version"],
    }

    uuid = b.get("@serialNumber")
    if uuid is not None:
        result["uuid"] = uuid

    try:
        hash_value = c["hashes"]["hash"]["#text"]
    except (TypeError, KeyError):
        pass
    else:
        result["hash"] = hash_value

    return result



# -*- coding: utf-8 -*-
"""Archivist connection interface

This module contains the base Archivist class which manages
the connection parameters to a DataTrails instance and
the basic REST verbs to GET, POST, PATCH and DELETE entities..

The REST methods in this class should only be used directly when
a CRUD endpoint for the specific type of entity is unavailable.
Current CRUD endpoints are assets, events, locations, attachments.
IAM subjects and IAM access policies.

Instantiation of this class encapsulates the URL and authentication
parameters (the max_time parameter is optional):

.. code-block:: python

   with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
       authtoken = tokenfile.read().strip()

   # Initialize connection to Archivist
   arch = Archivist(
       "https://app.datatrails.ai",
       authtoken,
       max_time=300.0,
   )

 The arch variable now has additional endpoints assets,events,locations,
 attachments, IAM subjects and IAM access policies documented elsewhere.

"""

from copy import deepcopy
from logging import getLogger
from time import time
from typing import Any, BinaryIO

from requests_toolbelt.multipart.encoder import MultipartEncoder

from .access_policies import _AccessPoliciesClient
from .appidp import _AppIDPClient
from .applications import _ApplicationsClient
from .archivistpublic import ArchivistPublic
from .assetattachments import _AssetAttachmentsClient
from .assets import _AssetsRestricted
from .attachments import _AttachmentsClient
from .compliance import _ComplianceClient
from .compliance_policies import _CompliancePoliciesClient
from .composite import _CompositeClient
from .confirmer import MAX_TIME
from .constants import (
    ROOT,
    SEP,
)
from .dictmerge import _dotdict
from .errors import (
    ArchivistError,
    _parse_response,
)
from .events import _EventsRestricted
from .locations import _LocationsClient
from .retry429 import retry_429
from .runner import _Runner
from .subjects import _SubjectsClient
from .tenancies import _TenanciesClient

LOGGER = getLogger(__name__)


class Archivist(ArchivistPublic):  # pylint: disable=too-many-instance-attributes
    """Base class for all Archivist endpoints.

    This class manages the connection to an Archivist instance and provides
    basic methods that represent the underlying REST interface.

    Args:
        url (str): URL of archivist endpoint
        auth: string representing JWT token, or a Tuple pair representing an
        Appregistration ID and secret.
        verify: if True the certificate is verified
        max_time (float): maximum time in seconds to wait for confirmation

    """

    # also change the type hints in __init__ below
    CLIENTS = {
        "access_policies": _AccessPoliciesClient,
        "assets": _AssetsRestricted,
        "assetattachments": _AssetAttachmentsClient,
        "appidp": _AppIDPClient,
        "applications": _ApplicationsClient,
        "attachments": _AttachmentsClient,
        "compliance": _ComplianceClient,
        "compliance_policies": _CompliancePoliciesClient,
        "composite": _CompositeClient,
        "events": _EventsRestricted,
        "locations": _LocationsClient,
        "runner": _Runner,
        "subjects": _SubjectsClient,
        "tenancies": _TenanciesClient,
    }

    def __init__(
        self,
        url: str,
        auth: "str|tuple[str,str]|None",
        *,
        fixtures: "dict[str,dict[Any,Any]]|None" = None,
        verify: bool = True,
        max_time: float = MAX_TIME,
        partner_id: str = "",
    ):
        super().__init__(
            fixtures=fixtures,
            verify=verify,
            max_time=max_time,
            partner_id=partner_id,
        )

        if isinstance(auth, tuple):
            self._machine_auth = auth
            self._auth = None
        else:
            self._auth = auth
            self._machine_auth = None

        self._expires_at = 0
        if url.endswith("/"):
            raise ArchivistError(f"URL {url} has trailing /")

        self._url = url
        self._root = SEP.join((url, ROOT))

        # Type hints for IDE autocomplete, keep in sync with CLIENTS map above
        self.access_policies: _AccessPoliciesClient
        self.appidp: _AppIDPClient
        self.applications: _ApplicationsClient
        self.assets: _AssetsRestricted
        self.assetattachments: _AssetAttachmentsClient
        self.attachments: _AttachmentsClient
        self.compliance: _ComplianceClient
        self.compliance_policies: _CompliancePoliciesClient
        self.composite: _CompositeClient
        self.events: _EventsRestricted
        self.locations: _LocationsClient
        self.runner: _Runner
        self.subjects: _SubjectsClient
        self.tenancies: _TenanciesClient

    def __str__(self) -> str:
        return f"Archivist({self._url})"

    def __getattr__(self, value: str) -> object:
        """Create endpoints on demand

        This only gets called when an atribute is not found.
        In this case the client attribute in question may not exist.
        """
        LOGGER.debug("getattr %s", value)
        client = self.CLIENTS.get(value)

        if client is None:
            raise AttributeError

        # set attribute so the method is no longer called for this
        # particular client
        c = client(self)
        super().__setattr__(value, c)
        return c

    @property
    def public(self) -> bool:
        """Not a public interface"""
        return False

    @property
    def url(self) -> str:
        """str: URL of Archivist endpoint"""
        return self._url

    @property
    def root(self) -> str:
        """str: ROOT of Archivist endpoint"""
        return self._root

    @property
    def auth(self) -> "str | None":
        """str: authorization token"""

        if self._auth is None and self._machine_auth is None:
            return None

        if self._machine_auth and self._expires_at < time():
            apptoken = self.appidp.token(*self._machine_auth)
            self._auth = apptoken.get("access_token")
            if self._auth is None:
                raise ArchivistError("Auth token from client id,secret is invalid")
            self._expires_at = time() + apptoken["expires_in"] - 10  # fudge factor
            LOGGER.info("Refresh token")

        return self._auth

    @property
    def Public(self) -> ArchivistPublic:  # pylint: disable=invalid-name
        """Get a Public instance"""
        arch = ArchivistPublic(
            fixtures=deepcopy(self._fixtures),
            verify=self._verify,
            max_time=self._max_time,
            partner_id=self._partner_id,
        )
        arch._user_agent = self._user_agent  # pylint: disable=protected-access
        return arch

    def __copy__(self) -> "Archivist":
        arch = Archivist(
            self._url,
            self.auth,
            fixtures=deepcopy(self._fixtures),
            verify=self._verify,
            max_time=self._max_time,
            partner_id=self._partner_id,
        )
        arch._user_agent = self._user_agent
        return arch

    def _add_headers(self, headers: "dict[str,str]|None") -> "dict[str,Any]":
        newheaders = super()._add_headers(headers)

        auth = self.auth  # this may trigger a refetch so only do it once here
        # for appidp endpoint there may not be an authtoken
        if auth is not None:
            newheaders["authorization"] = "Bearer " + auth.strip()

        return newheaders

    # currently only the archivist endpoint is allowed to create/modify data.
    # this may change...
    @retry_429
    def post(
        self,
        url: str,
        request: "dict[str,Any]|None",
        *,
        headers: "dict[str,Any]|None" = None,
        data: "dict[str, Any] | bool" = False,
    ) -> "dict[str, Any]":
        """POST method (REST)

        Creates an entity

        Args:
            url (str): e.g. v2/assets
            request (dict): request body defining the entity
            headers (dict): optional REST headers
            data (bool): send as form-encoded and not as json

        Returns:
            dict representing the response body (entity).
        """
        if data:
            response = self.session.post(
                url,
                data=request,
            )
        else:
            response = self.session.post(
                url,
                json=request,
                headers=self._add_headers(headers),
            )

        error = _parse_response(response)
        if error is not None:
            raise error

        return response.json()

    @retry_429
    def post_file(
        self,
        url: str,
        fd: BinaryIO,
        mtype: "str|None",
        *,
        form: str = "file",
        params: "dict[str, Any]|None" = None,
    ) -> "dict[str, Any]":
        """POST method (REST) - upload binary

        Uploads a file to an endpoint

        Args:
            url (str): e.g. v2/assets
            fd : iterable representing the contents of a file.
            mtype (str): mime type e.g. image/jpg
            params (dict): dictionary of optional path params

        Returns:
            dict representing the response body (entity).
        """
        multipart = MultipartEncoder(
            fields={
                form: ("filename", fd, mtype),
            }
        )
        headers = {
            "content-type": multipart.content_type,
        }

        response = self.session.post(
            url,
            data=multipart,  # pyright: ignore    https://github.com/requests/toolbelt/issues/312
            headers=self._add_headers(headers),
            params=_dotdict(params),
        )

        self._response_ring_buffer.appendleft(response)

        error = _parse_response(response)
        if error is not None:
            raise error

        return response.json()

    @retry_429
    def delete(
        self, url: str, *, headers: "dict[str, Any]|None" = None
    ) -> "dict[str, Any]":
        """DELETE method (REST)

        Deletes an entity

        Args:
            url (str): e.g. v2/assets/xxxxxxxxxxxxxxxxxxxxxxxxxxxx`
            headers (dict): optional REST headers

        Returns:
            dict representing the response body (entity).
        """
        response = self.session.delete(
            url,
            headers=self._add_headers(headers),
        )

        self._response_ring_buffer.appendleft(response)

        error = _parse_response(response)
        if error is not None:
            raise error

        return response.json()

    @retry_429
    def patch(
        self,
        url: str,
        request: "dict[str, Any]",
        *,
        headers: "dict[str, Any]| None" = None,
    ) -> "dict[str, Any]":
        """PATCH method (REST)

        Updates the specified entity.

        Args:
            url (str): e.g. v2/assets/xxxxxxxxxxxxxxxxxxxxxxxxxxxx`
            request (dict): request body defining the entity changes.
            headers (dict): optional REST headers

        Returns:
            dict representing the response body (entity).
        """

        response = self.session.patch(
            url,
            json=request,
            headers=self._add_headers(headers),
        )

        self._response_ring_buffer.appendleft(response)

        error = _parse_response(response)
        if error is not None:
            raise error

        return response.json()



"""Access_Policies interface

   Access to the access_policies endpoint.

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      access_policy = arch.access_policies.create(...)

"""

from copy import deepcopy
from logging import getLogger
from typing import TYPE_CHECKING, Any, Generator

# pylint:disable=cyclic-import      # but pylint doesn't understand this feature

if TYPE_CHECKING:
    from .archivist import Archivist

from .assets import Asset
from .constants import (
    ACCESS_POLICIES_LABEL,
    ACCESS_POLICIES_SUBPATH,
    ASSETS_LABEL,
)
from .dictmerge import _deepmerge

LOGGER = getLogger(__name__)


class AccessPolicy(dict):
    """AccessPolicy object"""

    @property
    def name(self) -> "str | None":
        """str: name of the access policy"""
        return self.get("display_name")


class _AccessPoliciesClient:
    """AccessPoliciesClient

    Access to access_policies entities using CRUD interface. This class is usually
    accessed as an attribute of the Archivist class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._subpath = f"{archivist_instance.root}/{ACCESS_POLICIES_SUBPATH}"
        self._label = f"{self._subpath}/{ACCESS_POLICIES_LABEL}"

    def __str__(self) -> str:
        return f"AccessPoliciesClient({self._archivist.url})"

    def create(
        self,
        props: "dict[str, Any]",
        filters: "list[dict[str, Any]]",
        access_permissions: "list[dict[str, Any]]",
    ) -> AccessPolicy:
        """Create access policy

        Creates access policy with defined attributes.

        Args:
            props (dict): properties of created access policy.
            filters (list): assets filters
            access permissions (list): list of access permissions

        Returns:
            :class:`AccessPolicy` instance

        """
        LOGGER.debug("Create Access Policy %s", props)
        return self.create_from_data(
            self.__params(
                props, filters=filters, access_permissions=access_permissions
            ),
        )

    def create_from_data(self, data: "dict[str, Any]") -> AccessPolicy:
        """Create access policy

        Creates access policy with request body from data stream.
        Suitable for reading data from a file using json.load or yaml.load

        Args:
            data (dict): request body of access policy.

        Returns:
            :class:`AccessPolicy` instance

        """
        return AccessPolicy(
            **self._archivist.post(
                f"{self._subpath}/{ACCESS_POLICIES_LABEL}",
                data,
            )
        )

    def read(self, identity: str) -> AccessPolicy:
        """Read Access Policy

        Reads access policy.

        Args:
            identity (str): access_policies identity e.g. access_policies/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            :class:`AccessPolicy` instance

        """
        return AccessPolicy(**self._archivist.get(f"{self._subpath}/{identity}"))

    def update(
        self,
        identity,
        *,
        props: "dict[str, Any] | None " = None,
        filters: "list[dict] | None " = None,
        access_permissions: "list[dict] | None " = None,
    ) -> AccessPolicy:
        """Update Access Policy

        Update access policy.

        Args:
            identity (str): access_policies identity e.g. access_policies/xxxxxxxxxxxxxxxxxxxxxxx
            props (dict): properties of created access policy.
            filters (list): assets filters
            access permissions (list): list of access permissions

        Returns:
            :class:`AccessPolicy` instance

        """
        return AccessPolicy(
            **self._archivist.patch(
                f"{self._subpath}/{identity}",
                self.__params(
                    props, filters=filters, access_permissions=access_permissions
                ),
            )
        )

    def delete(self, identity: str) -> "dict[str, Any]":
        """Delete Access Policy

        Deletes access policy.

        Args:
            identity (str): access_policies identity e.g. access_policies/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            :class:`AccessPolicy` instance - empty?

        """
        return self._archivist.delete(f"{self._subpath}/{identity}")

    def __params(
        self,
        props: "dict[str, Any] | None",
        *,
        filters: "list[dict] | None" = None,
        access_permissions: "list[dict] | None" = None,
    ) -> "dict[str, Any]":
        params = deepcopy(props) if props else {}
        if filters is not None:
            params["filters"] = filters

        if access_permissions is not None:
            params["access_permissions"] = access_permissions

        return _deepmerge(self._archivist.fixtures.get(ACCESS_POLICIES_LABEL), params)

    def count(self, *, display_name: "str | None" = None) -> int:
        """Count access policies.

        Counts number of access policies that match criteria.

        Args:
            display_name (str): display name (optional0

        Returns:
            integer count of access policies.

        """
        params = {"display_name": display_name} if display_name is not None else None
        return self._archivist.count(self._label, params=params)

    def list(
        self, *, page_size: "int|None" = None, display_name: "str|None" = None
    ) -> Generator[AccessPolicy, None, None]:
        """List access policies.

        List access policies that match criteria.

        Args:
            display_name (str): display name (optional0
            page_size (int): optional page size. (Rarely used).

        Returns:
            iterable that returns :class:`AccessPolicy` instances

        """
        params = {"display_name": display_name} if display_name is not None else None
        return (
            AccessPolicy(**a)
            for a in self._archivist.list(
                self._label,
                ACCESS_POLICIES_LABEL,
                page_size=page_size,
                params=params,
            )
        )

    # additional queries on different endpoints
    def list_matching_assets(
        self, access_policy_id: str, *, page_size: "int|None" = None
    ) -> Generator[Asset, None, None]:
        """List matching assets.

        List assets that match access policy.

        Args:
            access_policy_id (str): e.g. access_policies/xxxxxxxxxxxxxxx
            page_size (int): optional page size. (Rarely used).

        Returns:
            iterable that returns :class:`Asset` instances

        """
        return (
            Asset(**a)
            for a in self._archivist.list(
                f"{self._subpath}/{access_policy_id}/{ASSETS_LABEL}",
                ASSETS_LABEL,
                page_size=page_size,
            )
        )

    def list_matching_access_policies(
        self, asset_id: str, *, page_size: "int|None" = None
    ) -> Generator[AccessPolicy, None, None]:
        """List matching access policies.

        List access policies that match asset.

        Args:
            asset_id (str): e.g. assets/xxxxxxxxxxxxxxx
            page_size (int): optional page size. (Rarely used).

        Returns:
            iterable that returns :class:`AccessPolicy` instances

        """
        return (
            AccessPolicy(**a)
            for a in self._archivist.list(
                f"{self._subpath}/{asset_id}/{ACCESS_POLICIES_LABEL}",
                ACCESS_POLICIES_LABEL,
                page_size=page_size,
            )
        )



# -*- coding: utf-8 -*-
"""Public connection interface

   This module contains the base Archivist class which manages
   the public connection to an DataTrails instance and
   the basic REST verbs to GET, POST, PATCH and DELETE entities..

   The REST methods in this class should only be used directly when
   a CRUD endpoint for the specific type of entity is unavailable.
   Current CRUD endpoints are assets, events.
   Instantiation of this class encapsulates the URL and authentication
   parameters (the max_time parameter is optional):

   .. code-block:: python

      # Initialize connection to Archivist
      public = :Public(
          max_time=300.0,
      )

    The public variable now has additional endpoints assets,events.

"""

from collections import deque
from copy import deepcopy
from logging import getLogger
from typing import TYPE_CHECKING, Any, BinaryIO

import requests

if TYPE_CHECKING:
    from requests.models import Response


from .about import __version__ as VERSION
from .assetattachments import _AssetAttachmentsClient
from .assets import _AssetsPublic
from .confirmer import MAX_TIME
from .constants import (
    HEADERS_REQUEST_TOTAL_COUNT,
    HEADERS_TOTAL_COUNT,
    PARTNER_ID,
    USER_AGENT,
    USER_AGENT_PREFIX,
)
from .dictmerge import _deepmerge, _dotdict
from .errors import (
    ArchivistBadFieldError,
    ArchivistDuplicateError,
    ArchivistHeaderError,
    ArchivistNotFoundError,
    _parse_response,
)
from .events import _EventsPublic
from .headers import _headers_get
from .retry429 import retry_429

LOGGER = getLogger(__name__)


class ArchivistPublic:  # pylint: disable=too-many-instance-attributes
    """Base class for public Archivist endpoints.

    This class manages the connection to an Archivist instance and provides
    basic methods that represent the underlying REST interface.

    Args:
        verify: if True the certificate is verified
        max_time (float): maximum time in seconds to wait for confirmation

    """

    # also change the type hints in __init__ below
    CLIENTS = {
        "assets": _AssetsPublic,
        "events": _EventsPublic,
        "assetattachments": _AssetAttachmentsClient,
    }

    RING_BUFFER_MAX_LEN = 10

    def __init__(
        self,
        *,
        fixtures: "dict[str, Any]|None" = None,
        verify: bool = True,
        max_time: float = MAX_TIME,
        partner_id: str = "",
    ):
        self._verify = verify
        self._response_ring_buffer = deque(maxlen=self.RING_BUFFER_MAX_LEN)
        self._session = None
        self._max_time = max_time
        self._fixtures = fixtures or {}
        self._partner_id = partner_id
        self._user_agent = f"{USER_AGENT_PREFIX}{self.version}"

        # Type hints for IDE autocomplete, keep in sync with CLIENTS map above
        self.assets: _AssetsPublic
        self.events: _EventsPublic
        self.assetattachments: _AssetAttachmentsClient

    def __str__(self) -> str:
        return "ArchivistPublic()"

    def __getattr__(self, value: str) -> object:
        """Create endpoints on demand"""
        client = self.CLIENTS.get(value)

        if client is None:
            raise AttributeError

        c = client(self)
        super().__setattr__(value, c)
        return c

    def __enter__(self):
        """Just return self on entering - the session property will
        create the session when needed
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def session(self) -> requests.Session:
        """creates and returns session"""
        if self._session is None:
            self._session = requests.Session()
            self._session.verify = self.verify
        return self._session

    def close(self):
        """closes current session if open"""
        if self._session is not None:
            self._session.close()
            self._session = None

    @property
    def public(self) -> bool:
        """This is a public interface"""
        return True

    @property
    def root(self) -> str:
        """str: ROOT of Public endpoint"""
        return ""

    @property
    def verify(self) -> bool:
        """bool: Returns True if https connections are to be verified"""
        return self._verify

    @property
    def max_time(self) -> float:
        """bool: Returns maximum time in seconds to wait for confirmation"""
        return self._max_time

    @property
    def version(self) -> str:
        """str: Returns version of the archivist package"""
        return VERSION

    @property
    def partner_id(self) -> str:
        """str: Returns partner id if set when initialising an instance of this class"""
        return self._partner_id

    @property
    def user_agent(self) -> str:
        """str: Returns user agent"""
        return self._user_agent

    @user_agent.setter
    def user_agent(self, value):
        """str: Prepends user agent to user_agent property"""
        self._user_agent = f"{value} {self.user_agent}"

    @property
    def fixtures(self) -> "dict[str, Any]":
        """dict: Contains predefined attributes for each endpoint"""
        return self._fixtures

    @fixtures.setter
    def fixtures(self, fixtures: "dict[str, Any]"):
        """dict: Contains predefined attributes for each endpoint"""
        self._fixtures = _deepmerge(self._fixtures, fixtures)

    def __copy__(self):
        arch = ArchivistPublic(
            fixtures=deepcopy(self._fixtures),
            verify=self._verify,
            max_time=self._max_time,
            partner_id=self.partner_id,
        )
        arch._user_agent = self._user_agent
        return arch

    def _add_headers(self, headers: "dict[str, str]|None") -> "dict[str, str]":
        newheaders = {**headers} if headers is not None else {}
        newheaders[USER_AGENT] = self.user_agent

        p = self.partner_id
        if p:
            newheaders[PARTNER_ID] = p

        return newheaders

    # the public endpoint is currently readonly so only read-type methods are
    # defined here. This may change - the Public endpoint may allow writes
    # in future...
    @retry_429
    def get(
        self,
        url: str,
        *,
        headers: "dict[str, str]|None" = None,
        params: "dict[str, Any]|None" = None,
    ) -> "dict[str, Any]":
        """GET method (REST)

        Args:
            url (str): e.g. https://app.datatrails.ai/archivist/v2/publicassets/xxxxxxxxxxxxxxxxxx
            headers (dict): optional REST headers
            params (dict): optional params strings

        Returns:
            dict representing the response body (entity).

        """
        response = self.session.get(
            url,
            headers=self._add_headers(headers),
            params=_dotdict(params),
        )

        self._response_ring_buffer.appendleft(response)

        error = _parse_response(response)
        if error is not None:
            raise error

        return response.json()

    @retry_429
    def get_file(
        self,
        url: str,
        fd: BinaryIO,
        *,
        headers: "dict[str, str]|None" = None,
        params: "dict[str, Any]|None" = None,
    ) -> "Response":
        """GET method (REST) - chunked

        Downloads a binary object from upstream storage.

        Args:
            url (str): e.g. assets/xxxxxxxxxxxxxxxxxxxxxx
            identity (str): e.g. blobs/xxxxxxxxxxxxxxxxxxxxxxxxxxxx`
            fd (file): an iterable representing a file (usually from open())
                the file must be opened in binary mode
            headers (dict): optional REST headers
            params (dict): optional params strings

        Returns:
            REST response (not the response body)

        """
        response = self.session.get(
            url,
            headers=self._add_headers(headers),
            stream=True,
            params=_dotdict(params),
        )

        self._response_ring_buffer.appendleft(response)

        error = _parse_response(response)
        if error is not None:
            raise error

        for chunk in response.iter_content(chunk_size=4096):
            if chunk:
                fd.write(chunk)

        return response

    @retry_429
    def __list(
        self,
        url: str,
        params: "dict[str, Any]|None",
        *,
        page_size: "int|None" = None,
        headers: "dict[str, str]|None" = None,
    ) -> "Response":
        if page_size is not None:
            if params is not None:
                params["page_size"] = page_size
            else:
                params = {"page_size": page_size}

        response = self.session.get(
            url,
            headers=self._add_headers(headers),
            params=_dotdict(params),
        )

        self._response_ring_buffer.appendleft(response)

        error = _parse_response(response)
        if error is not None:
            raise error

        return response

    def last_response(self, *, responses: int = 1) -> "list[Response]":
        """Returns the requested number of response objects from the response ring buffer

        Args:
            responses (int): Number of responses to be returned in a list

        Returns:
            list of responses.

        """

        return list(self._response_ring_buffer)[:responses]

    def get_by_signature(
        self,
        url: str,
        field: str,
        params: "dict[str, Any]",
        *,
        headers: "dict[str, str]|None" = None,
    ) -> "dict[str, Any]":
        """GET method (REST) with params string

        Reads an entity indirectly by searching for its signature

        It is expected that the params parameters will result in only a single entity
        being found.

        Args:
            url (str): e.g. https://app.datatrails.ai/archivist/v2/assets
            field (str): name of collection of entities e.g assets
            params (dict): selector e.g. {"attributes": {"arc_display_name":"container no. 1"}}
            headers (dict): optional REST headers

        Returns:
            dict representing the entity found.

        Raises:
            ArchivistBadFieldError: field has incorrect value.
            ArchivistNotFoundError: No entity found
            ArchivistDuplicateError: More than one entity matching signature found

        """

        response = self.__list(
            url,
            params,
            page_size=2,
            headers=headers,
        )

        data = response.json()

        try:
            records = data[field]
        except KeyError as ex:
            raise ArchivistBadFieldError(f"No {field} found") from ex

        if len(records) == 0:
            raise ArchivistNotFoundError("No entity found")

        if len(records) > 1:
            raise ArchivistDuplicateError(f"{len(records)} found")

        return records[0]

    def count(self, url: str, *, params: "dict[str, Any]|None" = None) -> int:
        """GET method (REST) with params string

        Returns the count of objects that match params

        Args:
            url (str): e.g. https://app.datatrails.ai/archivist/v2/assets
            params (dict): selector e.g. {"attributes":{"arc_display_name":"container no. 1"}}

        Returns:
            integer count of entities found.

        Raises:
            ArchivistHeaderError: If the expected count header is not present

        """

        response = self.__list(
            url,
            params,
            page_size=1,
            headers={HEADERS_REQUEST_TOTAL_COUNT: "true"},
        )

        count = _headers_get(response.headers, HEADERS_TOTAL_COUNT)

        if count is None:
            raise ArchivistHeaderError("Did not get a count in the header")

        return int(count)

    def list(
        self,
        url: str,
        field: str,
        *,
        page_size: "int|None" = None,
        params: "dict[str, Any]|None" = None,
        headers: "dict[str, str]|None" = None,
    ):
        """GET method (REST) with params string

        Lists entities that match the params dictionary.

        If page size is specified return the list of records in batches of page_size
        until next_page_token in response is null.

        If page size is unspecified return up to the internal limit of records.
        (different for each endpoint)

        Args:
            url (str): e.g. https://app.datatrails.ai/archivist/v2/assets
            field (str): name of collection of entities e.g assets
            page_size (int): optional number of items per request e.g. 500
            params (dict): selector e.g. {"confirmation_status": "CONFIRMED", }
            headers (dict): optional REST headers

        Returns:
            iterable that lists entities

        Raises:
            ArchivistBadFieldError: field has incorrect value.

        """

        while True:
            response = self.__list(
                url,
                params,
                page_size=page_size,
                headers=headers,
            )
            data = response.json()

            try:
                records = data[field]
            except KeyError as ex:
                raise ArchivistBadFieldError(f"No {field} found") from ex

            yield from records

            page_token = data.get("next_page_token")
            if not page_token:
                break

            params = {"page_token": page_token}



"""assets confirmer interface
"""

from copy import deepcopy
from logging import getLogger
from typing import TYPE_CHECKING, Any, Union, overload

import backoff

if TYPE_CHECKING:
    # pylint:disable=cyclic-import      # but pylint doesn't understand this feature
    from backoff._typing import Details

    from .assets import Asset, _AssetsPublic, _AssetsRestricted
    from .events import Event, _EventsPublic, _EventsRestricted


from .confirmation_status import ConfirmationStatus
from .constants import CONFIRMATION_STATUS
from .errors import ArchivistUnconfirmedError
from .utils import backoff_handler

MAX_TIME = 300
LOGGER = getLogger(__name__)

# pylint: disable=protected-access
PublicManagers = Union["_AssetsPublic", "_EventsPublic"]
PrivateManagers = Union["_AssetsRestricted", "_EventsRestricted"]
Managers = Union[PublicManagers, PrivateManagers]
ReturnTypes = Union["Asset", "Event"]


def __lookup_max_time():
    return MAX_TIME


def __on_giveup_confirmation(details: "Details"):
    identity: str = details["args"][1]
    elapsed: float = details["elapsed"]
    raise ArchivistUnconfirmedError(
        f"confirmation for {identity} timed out after {elapsed} seconds"
    )


# These overloads are used for type hinting, if self is events client then
# an event will be returned. If self is Asset client then an asset will be
# returned. Overloads are evaluated at startup but not at runtime, therefore
# no test coverage be done directly.


@overload
def _wait_for_confirmation(
    self: "_AssetsRestricted",
    identity: str,
) -> "Asset": ...  # pragma: no cover


@overload
def _wait_for_confirmation(
    self: "_AssetsPublic",
    identity: str,
) -> "Asset": ...  # pragma: no cover


@overload
def _wait_for_confirmation(
    self: "_EventsRestricted",
    identity: str,
) -> "Event": ...  # pragma: no cover


@overload
def _wait_for_confirmation(
    self: "_EventsPublic",
    identity: str,
) -> "Event": ...  # pragma: no cover


@backoff.on_predicate(
    backoff.expo,
    max_value=30.0,
    logger=None,  # pyright: ignore
    max_time=__lookup_max_time,
    on_backoff=backoff_handler,
    on_giveup=__on_giveup_confirmation,
)
def _wait_for_confirmation(self: Managers, identity: str) -> ReturnTypes:
    """Return None until entity is confirmed"""

    entity = self.read(identity)

    LOGGER.debug("entity %s", entity)
    if CONFIRMATION_STATUS not in entity:
        raise ArchivistUnconfirmedError(
            f"cannot confirm {identity} as confirmation_status is not present"
        )

    status = entity[CONFIRMATION_STATUS]
    if status == ConfirmationStatus.FAILED.name:
        raise ArchivistUnconfirmedError(
            f"confirmation for {identity} FAILED - this is unusable"
        )

    if status in (
        ConfirmationStatus.CONFIRMED.name,
        ConfirmationStatus.COMMITTED.name,
        ConfirmationStatus.UNEQUIVOCAL.name,
    ):
        return entity

    return None  # pyright: ignore


def __on_giveup_confirmed(details: "Details"):
    self: PrivateManagers = details["args"][0]
    count = self.pending_count
    elapsed: float = details["elapsed"]
    raise ArchivistUnconfirmedError(
        f"{count} pending assets still present after {elapsed} seconds"
    )


@backoff.on_predicate(
    backoff.expo,
    max_value=30.0,
    logger=None,  # pyright: ignore
    max_time=__lookup_max_time,
    on_backoff=backoff_handler,
    on_giveup=__on_giveup_confirmed,
)
def _wait_for_confirmed(
    self: PrivateManagers,
    *,
    props: "dict[str, Any]|None" = None,
    **kwargs: Any,
) -> bool:
    """Return False until all entities are confirmed"""

    # look for pending entities
    newprops = deepcopy(props) if props else {}
    newprops[CONFIRMATION_STATUS] = ConfirmationStatus.PENDING.name
    LOGGER.debug("Count pending entities %s", newprops)
    pending_count = self.count(props=newprops, **kwargs)

    # look for stored entities
    newprops = deepcopy(props) if props else {}
    newprops[CONFIRMATION_STATUS] = ConfirmationStatus.STORED.name
    LOGGER.debug("Count stored entities %s", newprops)
    stored_count = self.count(props=newprops, **kwargs)

    count = pending_count + stored_count

    if count == 0:
        # did any fail
        newprops = deepcopy(props) if props else {}
        newprops[CONFIRMATION_STATUS] = ConfirmationStatus.FAILED.name
        count = self.count(props=newprops, **kwargs)
        if count > 0:
            raise ArchivistUnconfirmedError(f"There are {count} FAILED entities")

        return True

    self.pending_count = count
    return False



"""Assets interface

   Access to the assets endpoint.

   The user is not expected to use this class directly. It is an attribute of the
   :class:`Archivist` class.

   For example instantiate an Archivist instance and execute the methods of the class:

   .. code-block:: python

      with open(".auth_token", mode="r", encoding="utf-8") as tokenfile:
          authtoken = tokenfile.read().strip()

      # Initialize connection to Archivist
      arch = Archivist(
          "https://app.datatrails.ai",
          authtoken,
      )
      asset = arch.assets.create(...)

"""

from copy import deepcopy
from logging import getLogger
from typing import TYPE_CHECKING, Any

# pylint:disable=cyclic-import      # but pylint doesn't understand this feature
from . import confirmer
from .asset import Asset
from .constants import (
    ASSET_BEHAVIOURS,
    ASSETS_LABEL,
    ASSETS_SUBPATH,
    CONFIRMATION_STATUS,
)
from .dictmerge import _deepmerge
from .errors import ArchivistBadFieldError, ArchivistNotFoundError
from .utils import selector_signature

if TYPE_CHECKING:
    from .archivist import Archivist

LOGGER = getLogger(__name__)


class _AssetsPublic:
    """AssetsReader

    Access to assets entities using CRUD interface. This class is usually
    accessed as an attribute of the Archivist or Public class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        self._archivist = archivist_instance
        self._public = archivist_instance.public
        self._subpath = f"{archivist_instance.root}/{ASSETS_SUBPATH}"

    def __str__(self) -> str:
        return "AssetsPublic()"

    def _identity(self, identity: str) -> str:
        """Return fully qualified identity
        If public then expect a full url as argument
        """
        if self._public:
            return identity

        return f"{self._subpath}/{identity}"

    def read(self, identity: str) -> Asset:
        """Read asset

        Reads asset.

        Args:
            identity (str): assets identity e.g. assets/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            :class:`Asset` instance

        """
        return Asset(**self._archivist.get(self._identity(identity)))


class _AssetsRestricted(_AssetsPublic):
    """AssetsRestricted

    Access to assets entities using CRUD interface. This class is usually
    accessed as an attribute of the Archivist or Public class.

    Args:
        archivist (Archivist): :class:`Archivist` instance

    """

    def __init__(self, archivist_instance: "Archivist"):
        super().__init__(archivist_instance)
        self._label = f"{self._subpath}/{ASSETS_LABEL}"
        self.pending_count: int = 0

    def __str__(self) -> str:
        return f"AssetsRestricted({self._archivist.url})"

    def __params(
        self, props: "dict[str, Any]|None", attrs: "dict[str, Any]|None"
    ) -> "dict[str, Any]":
        params = deepcopy(props) if props else {}
        if attrs:
            params["attributes"] = attrs

        return _deepmerge(self._archivist.fixtures.get(f"{ASSETS_LABEL}"), params)

    def create(
        self,
        *,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
        confirm: bool = False,
    ) -> Asset:
        """Create asset

        Creates asset with defined properties and attributes.

        Args:
            props (dict): Properties
            attrs (dict): attributes of created asset.
            confirm (bool): if True wait for asset to be confirmed.

        Returns:
            :class:`Asset` instance

        """
        LOGGER.debug("Create Asset %s", attrs)
        # default behaviours  are added first - any set in user-specified fixtures or
        # in the method args will overide...
        newprops = _deepmerge({"behaviours": ASSET_BEHAVIOURS}, props)
        data = self.__params(newprops, attrs)
        return self.create_from_data(data, confirm=confirm)

    def create_from_data(
        self, data: "dict[str, Any]", *, confirm: bool = False
    ) -> Asset:
        """Create asset

        Creates asset with request body from data stream.
        Suitable for reading data from a file using json.load or yaml.load

        Args:
            data (dict): request body of asset.
            confirm (bool): if True wait for asset to be confirmed.

        Returns:
            :class:`Asset` instance

        """
        asset = Asset(**self._archivist.post(self._label, data))
        if not confirm:
            return asset

        return self.wait_for_confirmation(asset["identity"])

    def create_if_not_exists(
        self, data: "dict[str, Any]", *, confirm: bool = False
    ) -> "tuple[Asset, bool]":
        """
        Creates an asset and associated locations and attachments if asset
        does not already exist.

        Args:
            data (dict): request body of asset.
            confirm (bool): if True wait for asset to be confirmed.

        A YAML representation of the data argument would be:

            .. code-block:: yaml

                selector:
                  - attributes:
                    - arc_display_name
                behaviours
                  - RecordEvidence
                attributes:
                  arc_display_name: DataTrails Front Door
                  arc_firmware_version: "1.0"
                  arc_serial_number: das-j1-01
                  arc_description: Electronic door entry system to DataTrails France
                  wavestone_asset_id: paris.france.datatrails.das
                location:
                  identity: locations/xxxxxxxxxxxxxxxxxxxxxxxxxx
                location:
                  selector:
                    - display_name
                  display_name: DataTrails Paris
                  description: Sales and sales support for the French region
                  latitude: 48.8339211,
                  longitude: 2.371345,
                  attributes:
                    address: 5 Parvis Alan Turing, 75013 Paris, France
                    wavestone_ext: managed
                attachments:
                  - filename: functests/test_resources/doors/assets/entry-terminal.jpg
                    content_type: image/jpg
                    attachment: terminal entry

            The 'selector' value is required and will usually specify the 'arc_display_name' as a
            secondary key. The keys in 'selector' must exist in the attributes of the asset.

            If 'location' is specified then the 'selector' value is required and is used as a
            secondary key. Likewise the secondary key must exist in the attributes of the location.

            Alternatively the identity of the location is specified - both
            are shown - choose one.

        Returns:
            tuple of :class:`Asset` instance, Boolean is True if asset already existed

        """

        asset = None
        existed = False
        data = deepcopy(data)
        attachments = data.pop("attachments", None)
        location = data.pop("location", None)
        selector = data.pop("selector")  # must exist
        props, attrs = selector_signature(selector, data)
        try:
            asset = self.read_by_signature(props=props, attrs=attrs)

        except ArchivistNotFoundError:
            LOGGER.info(
                "asset with selector %s,%s does not exist - creating", props, attrs
            )

        else:
            LOGGER.info("asset with selector %s,%s already exists", props, attrs)
            return asset, True

        # is location present?
        if location is not None:
            if "identity" in location:
                data["attributes"]["arc_home_location_identity"] = location["identity"]
            else:
                loc, _ = self._archivist.locations.create_if_not_exists(
                    location,
                )
                data["attributes"]["arc_home_location_identity"] = loc["identity"]

        # any attachments ?
        if attachments is not None:
            for a in attachments:
                # attempt to get attachment to use as a key
                attachment_key = a.get("attachment", None)
                if attachment_key is None:
                    # failing that create a key from filename or url
                    attachment_key = self._archivist.attachments.get_default_key(a)
                data["attributes"][attachment_key] = self._archivist.attachments.create(
                    a
                )

        asset = self.create_from_data(
            data=data,
            confirm=confirm,
        )

        return asset, existed

    def wait_for_confirmation(self, identity: str) -> Asset:
        """Wait for asset to be confirmed.

        Waits for asset to be confirmed.

        Args:
            identity (str): identity of asset

        Returns:
            True if asset is confirmed.

        """
        confirmer.MAX_TIME = self._archivist.max_time
        # pylint: disable=protected-access
        return confirmer._wait_for_confirmation(self, identity)

    def wait_for_confirmed(
        self,
        *,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
    ) -> bool:
        """Wait for assets to be confirmed.

        Waits for all assets that match criteria to be confirmed.

        Args:
            props (dict): e.g. {"tracked": "TRACKED" }
            attrs (dict): e.g. {"arc_display_type": "door" }

        Returns:
            True if all assets are confirmed.

        """
        # check that entities exist
        newprops = deepcopy(props) if props else {}
        newprops.pop(CONFIRMATION_STATUS, None)

        LOGGER.debug("Count assets %s", newprops)
        count = self.count(props=newprops, attrs=attrs)
        if count == 0:
            raise ArchivistNotFoundError("No assets exist")

        confirmer.MAX_TIME = self._archivist.max_time
        # pylint: disable=protected-access
        return confirmer._wait_for_confirmed(self, props=newprops, attrs=attrs)

    def count(
        self,
        *,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
    ) -> int:
        """Count assets.

        Counts number of assets that match criteria.

        Args:
            props (dict): e.g. {"confirmation_status": "CONFIRMED" }
            attrs (dict): e.g. {"arc_display_type": "door" }

        Returns:
            integer count of assets.

        """
        return self._archivist.count(self._label, params=self.__params(props, attrs))

    def list(
        self,
        *,
        page_size: "int|None" = None,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
    ):
        """List assets.

        Lists assets that match criteria.

        Args:
            props (dict): optional e.g. {"tracked": "TRACKED" }
            attrs (dict): optional e.g. {"arc_display_type": "door" }
            page_size (int): optional page size. (Rarely used).

        Returns:
            iterable that returns :class:`Asset` instances

        """
        return (
            Asset(**a)
            for a in self._archivist.list(
                self._label,
                ASSETS_LABEL,
                page_size=page_size,
                params=self.__params(props, attrs),
            )
        )

    def read_by_signature(
        self,
        *,
        props: "dict[str, Any]|None" = None,
        attrs: "dict[str, Any]|None" = None,
    ) -> Asset:
        """Read Asset by signature.

        Reads asset that meets criteria. Only one asset is expected.

        Args:
            props (dict): e.g. {"tracked": "TRACKED" }
            attrs (dict): e.g. {"arc_display_type": "door" }

        Returns:
            :class:`Asset` instance

        """
        assets_label = f"public{ASSETS_LABEL}" if self._public else ASSETS_LABEL
        return Asset(
            **self._archivist.get_by_signature(
                self._label,
                assets_label,
                params=self.__params(props, attrs),
            )
        )

    def publicurl(self, identity: str) -> str:
        """Read asset public url

        Reads assets public url.

        Args:
            identity (str): assets identity e.g. assets/xxxxxxxxxxxxxxxxxxxxxxx

        Returns:
            :str: publicurl as string

        """
        body = self._archivist.get(f"{self._identity(identity)}:publicurl")
        publicurl = body.get("publicurl")
        if publicurl is None:
            raise ArchivistBadFieldError("No publicurl found in response")

        return publicurl
